# -*- coding: utf-8 -*-

import sympy
import numpy as np
import pandas as pd
from itertools import product
from scipy.linalg import block_diag

sympy.init_printing()
from sympy import symbols

from dramkit.gentools import link_lists


#%%
def get_matrixs_by_sasr(pi_s2a,
                        p_sa2sr,
                        rewards,
                        s_names=None,
                        a_names=None):
    '''
    | 根据策略矩阵pi_s2a: P(s,a)和动力系统矩阵p_sa2sr: P(s',r'|s,a)
    | 生成相关矩阵
    '''
    
    pi_s2a, p_sa2sr = np.array(pi_s2a), np.array(p_sa2sr)
    rewards = list(rewards)
    n_s, n_a = pi_s2a.shape
    n_sa, n_sr = p_sa2sr.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))
    sr_names = list(product(s_names, rewards))        
    
    # 策略状态-动作概率矩阵
    pi_s2a = pd.DataFrame(pi_s2a, index=s_names, columns=a_names)
    pi_s2a.index.name = 's'
    pi_s2a.columns.name = 'a'
    # 状态&动作-状态&奖励概率矩阵
    p_sa2sr = pd.DataFrame(p_sa2sr, index=sa_names, columns=sr_names)
    p_sa2sr.index.name = 'sa'
    p_sa2sr.columns.name = 'sr'
    
    # 状态&动作-状态转移概率矩阵
    p_sa2s = p_sa2sr.transpose()
    p_sa2s = p_sa2s.reset_index()
    p_sa2s[['s', 'r']] = p_sa2s['sr'].apply(pd.Series)
    def _sum(df):
        return df.drop(['sr', 's', 'r'], axis=1).sum()
    p_sa2s = p_sa2s.groupby('s').apply(lambda x: _sum(x))
    p_sa2s = p_sa2s.transpose()
    p_sa2s = p_sa2s.reindex(columns=s_names,
                            index=p_sa2sr.index)
    
    # 策略状态-动作矩阵构造分块对角矩阵
    pi_s2a_diag = block_diag(*(pi_s2a.iloc[k, :] for k in range(pi_s2a.shape[0])))
    
    # 策略下状态-状态转移概率矩阵
    pi_s2s = np.matmul(pi_s2a_diag, p_sa2s)
    pi_s2s.index = s_names
    pi_s2s.index.name = 's'
    
    # 策略下状态&动作-状态&动作转移概率矩阵
    pi_sa2sa = pd.concat([p_sa2s]*len(a_names), axis=1)
    pi_sa2sa.columns = [(s, a) for a in a_names for s in p_sa2s.columns]
    pi_sa2sa = pi_sa2sa.reindex(index=p_sa2s.index,
                                columns=p_sa2s.index)
    pi_sa2sa = pi_sa2sa * link_lists(pi_s2a.to_dict('split')['data'])
    
    r_e_sa2s = [x[1] for x in sr_names]
    r_e_sa2s = r_e_sa2s * p_sa2sr
    # 状态&动作期望奖励
    r_e_sa = r_e_sa2s.sum(axis=1).to_frame()
    r_e_sa.columns = ['r']
    # 状态期望奖励
    r_e_s = r_e_sa.reset_index()
    r_e_s[['s', 'a']] = r_e_s['sa'].apply(pd.Series)
    def _sum2(df):
        return df.drop(['sa', 's', 'a'], axis=1).sum()
    r_e_s = r_e_s.groupby('s').apply(lambda x: _sum2(x))
    r_e_s = r_e_s.reindex(index=pi_s2a.index)
    # 状态&动作-状态期望奖励
    r_e_sa2s = r_e_sa2s.transpose()
    r_e_sa2s = r_e_sa2s.reset_index()
    r_e_sa2s[['s', 'r']] = r_e_sa2s['sr'].apply(pd.Series)
    def _sum3(df):
        return df.drop(['sr', 's', 'r'], axis=1).sum()
    r_e_sa2s = r_e_sa2s.groupby('s').apply(lambda x: _sum3(x))
    r_e_sa2s = r_e_sa2s.transpose()
    r_e_sa2s = r_e_sa2s.reindex(columns=s_names,
                            index=p_sa2s.index)
    # 状态&动作-状态奖励
    r_sa2s = r_e_sa2s / p_sa2s.replace(0, np.nan) # 概率为0时计算不出来
    
    # 策略下每个状态的期望奖励
    r_e_pi_s = pi_s2a * r_e_sa.values.reshape(n_s, n_a)
    r_e_pi_s = r_e_pi_s.sum(axis=1).to_frame()
    r_e_pi_s.columns = ['r']
    
    return  pi_s2a, p_sa2sr, p_sa2s, pi_s2a_diag, pi_s2s, \
            pi_sa2sa, r_sa2s, r_e_sa2s, r_e_sa, r_e_s, r_e_pi_s
            
            
def get_matrixs_by_sas(pi_s2a,
                       p_sa2s,
                       r_sa2s,
                       s_names=None,
                       a_names=None):
    '''
    | 根据策略矩阵pi_s2a: P(s,a)和动力系统矩阵p_sa2s: P(s'|s,a)、r_sa2s: R(s'|s,a)
    | 生成相关矩阵
    '''
    
    pi_s2a, p_sa2s = np.array(pi_s2a), np.array(p_sa2s)
    n_s, n_a = pi_s2a.shape
    n_sa, n_s = p_sa2s.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))        
    
    # 策略状态-动作概率矩阵
    pi_s2a = pd.DataFrame(pi_s2a, index=s_names, columns=a_names)
    pi_s2a.index.name = 's'
    pi_s2a.columns.name = 'a'
    # 状态&动作-状态转移概率矩阵
    p_sa2s = pd.DataFrame(p_sa2s, index=sa_names, columns=s_names)
    p_sa2s.index.name = 'sa'
    p_sa2s.columns.name = 's'
    # 状态&动作-状态奖励
    r_sa2s = pd.DataFrame(r_sa2s, index=sa_names, columns=s_names)
    r_sa2s.index.name = 'sa'
    r_sa2s.columns.name = 's'
        
    # 策略状态-动作矩阵构造分块对角矩阵
    pi_s2a_diag = block_diag(*(pi_s2a.iloc[k, :] for k in range(pi_s2a.shape[0])))
    
    # 策略下状态-状态转移概率矩阵
    pi_s2s = np.matmul(pi_s2a_diag, p_sa2s)
    pi_s2s.index = s_names
    pi_s2s.index.name = 's'
    
    # 策略下状态&动作-状态&动作转移概率矩阵
    pi_sa2sa = pd.concat([p_sa2s]*len(a_names), axis=1)
    pi_sa2sa.columns = [(s, a) for a in a_names for s in p_sa2s.columns]
    pi_sa2sa = pi_sa2sa.reindex(index=p_sa2s.index,
                                columns=p_sa2s.index)
    pi_sa2sa = pi_sa2sa * link_lists(pi_s2a.to_dict('split')['data'])
    
    # 状态&动作-状态期望奖励
    r_e_sa2s = r_sa2s * p_sa2s
    
    # 状态&动作期望奖励
    r_e_sa = r_e_sa2s.sum(axis=1).to_frame()
    r_e_sa.columns = ['r']
    
    # 状态期望奖励
    r_e_s = r_e_sa.reset_index()
    r_e_s[['s', 'a']] = r_e_s['sa'].apply(pd.Series)
    def _sum2(df):
        return df.drop(['sa', 's', 'a'], axis=1).sum()
    r_e_s = r_e_s.groupby('s').apply(lambda x: _sum2(x))
    r_e_s = r_e_s.reindex(index=pi_s2a.index)
    
    # 策略下每个状态的期望奖励
    r_e_pi_s = pi_s2a * r_e_sa.values.reshape(n_s, n_a)
    r_e_pi_s = r_e_pi_s.sum(axis=1).to_frame()
    r_e_pi_s.columns = ['r']
    
    p_sa2sr = None
    
    return  pi_s2a, p_sa2sr, p_sa2s, pi_s2a_diag, pi_s2s, \
            pi_sa2sa, r_sa2s, r_e_sa2s, r_e_sa, r_e_s, r_e_pi_s
            
#%%
def bellman_sympy_sasr(pi_s2a, p_sa2sr, rewards, gamma,
                       s_names=None, a_names=None):
    pi_s2a, p_sa2sr, p_sa2s, pi_s2a_diag, pi_s2s, pi_sa2sa, \
    r_sa2s, r_e_sa2s, r_e_sa, r_e_s, r_e_pi_s = \
            get_matrixs_by_sasr(pi_s2a, p_sa2sr, rewards,
                                s_names=s_names, a_names=a_names)
    results, system = {}, None
    return results, system

#%%
if __name__ == '__main__':
    from IPython.display import display
    from dramkit.gentools import TimeRecoder
    
    tr = TimeRecoder()
    
    #%%
    # 饥饿游戏
    s_names = ['Hgry', 'Full'] # 状态空间
    a_names = ['notEat', 'Eat'] # 动作空间
    rewards = [-3, -2, 1, 2] # 奖励空间
    x, y = symbols('x y')
    alpha, beta, gamma = symbols('alpha beta gamma')
    # 策略概率表
    pi_s2a = [[1-x,   x],
              [  y, 1-y]]
    # 状态&动作-状态&奖励概率表
    p_sa2sr = [[      0,    1, 0, 0, 0, 0,     0,      0],
               [1-alpha,    0, 0, 0, 0, 0, alpha,      0],
               [      0, beta, 0, 0, 0, 0,     0, 1-beta],
               [      0,    0, 0, 0, 0, 0,     1,      0]]
    # 状态&动作-状态转移概率表
    p_sa2s = [[      1,      0],
              [1-alpha,  alpha],
              [   beta, 1-beta],
              [      0,      1]]
    # 状态&动作-状态奖励表
    r_sa2s = [[-2, 0],
              [-3, 1],
              [-2, 2],
              [ 0, 1]]
    
    # '''
    # 1. 使用带奖励的状态&动作-状态&奖励转移矩阵---------------------------
    res1, sys1 = bellman_sympy_sasr(
                    pi_s2a, p_sa2sr, rewards, gamma,
                    s_names=s_names, a_names=a_names)
    for k, v in res1.items():
        display(k, v)
    # '''
    
    '''
    # alpha_, beta_, gamma_ = 0.67, 0.75, 0.80
    alpha_, beta_, gamma_ = 2/3, 3/4, 4/5    
        
    # 2. 使用状态&动作-状态转移矩阵
    # '''
           
    '''    
    # 3. 仅对状态价值自身相互表示
    res3, (sys3, ynames3, p_ss3, r_s3, p_sa3, p_sas3, r_sas3) = \
            bellman_sas_sv(pi_sa, p_sas, r_sas, gamma,
                           s_names=s_names, a_names=a_names)
    res3_ = res3.copy()
    display(res3)
    for k, v in res3.items():        
        display(k, v)
        v_subs = v.subs([(alpha, alpha_), 
                         (beta, beta_),
                         (gamma, gamma_)])
        display(k, v_subs)
        
        
    # 4. 仅对动作价值自身相互表示
    res4, (sys4, ynames4, p_sasa4, r_sa4) = \
            bellman_sas_av(pi_sa, p_sas, r_sas, gamma,
                           s_names=s_names, a_names=a_names)
    display(res4)
    for k, v in res4.items():        
        display(k, v)
        display(k, v.subs([(alpha, alpha_), 
                            (beta, beta_),
                            (gamma, gamma_)]))
        
        
    # 求状态价值最大值
    xv = yv = np.linspace(0, 1, 50)
    dfv = pd.DataFrame(product(xv, yv), columns=['x', 'y'])
    dfv['x'] = 1 - dfv['x']
    for k, v in res3_.items():
        v_name = str(k)
        dfv[v_name] = dfv[['x', 'y']].apply(
                    lambda xy: v.subs(
                        [(alpha, alpha_), (beta, beta_),
                         (gamma, gamma_),
                         (x, xy['x']), (y, xy['y'])]),
                    axis=1)
    print(dfv.max())
    # '''
    
    #%%
    tr.used()
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    