# -*- coding: utf-8 -*-

import sympy
import numpy as np
import pandas as pd
sympy.init_printing()
from sympy import symbols
from itertools import product
from scipy.linalg import block_diag
from dramkit.gentools import df_rows2cols, df_cols2rows

#%%
def bellman_sasr(p_sa, p_sasr, rewards, gamma,
                 s_names=None, a_names=None,
                 solve=True):
    '''
    | 贝尔曼方程矩阵形式并求解，动力系统矩阵p_sasr为P(s',r'|s,a)形式
    | p_sa为策略概率矩阵
    | p_sasr为带奖励值的状态&动作-状态转移概率矩阵
    | rewards为奖励空间列表
    | gamma为奖励的权重衰减系数（折扣因子）
    | s_names和a_names分别为状态空间和动作空间的名称列表
    | 注意：概率矩阵、奖励空间列表和状态空间及动作空间名称的顺序必须是对应的
    '''
    
    p_sa, p_sasr = np.array(p_sa), np.array(p_sasr)
    rewards = list(rewards)
    n_s, n_a = p_sa.shape
    n_sa, n_sr = p_sasr.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))
    sr_names = list(product(s_names, rewards))        
    
    p_sa = pd.DataFrame(p_sa, index=s_names, columns=a_names)
    p_sasr = pd.DataFrame(p_sasr, index=sa_names, columns=sr_names)
    
    part1 = np.eye(n_s)
    part2 = block_diag(*(p_sa.iloc[k, :] for k in range(p_sa.shape[0])))
    part12 = np.concatenate([part1, -part2], axis=1)
    
    part3 = -gamma * p_sasr
    for s in s_names:
        part3[s] = part3[[x for x in part3.columns if s == x[0]]].sum(axis=1)
    part3 = part3.reindex(columns=s_names)
    part4 = np.eye(n_sa)
    part34 = np.concatenate([part3, part4], axis=1)
    
    a_ = np.concatenate([part12, part34], axis=0)
    
    y1 = [[0] for _ in range(n_s)]
    y2 = [x[1] for x in sr_names]
    y2 = y2 * p_sasr
    # y2_ = [x[1] for x in sr_names]
    # y2 = p_sasr.copy()
    # for k in range(len(y2_)):
    #     y2[y2.columns[k]] = y2_[k] * y2[y2.columns[k]]
    y2 = y2.sum(axis=1)
    y_ = np.concatenate((y1, pd.DataFrame(y2)), axis=0)
    
    system = np.concatenate([a_, y_], axis=1)
    system = sympy.Matrix(system)
    
    ynames = []
    for s in s_names:
        exec('v_{s} = symbols("v_{s}")'.format(s=s))
        ynames.append(eval('v_{}'.format(s)))
    for sa in sa_names:
        exec('v_{sa0}_{sa1} = symbols("v_{sa0}_{sa1}")'.format(sa0=sa[0], sa1=sa[1]))
        ynames.append(eval('v_{}_{}'.format(sa[0], sa[1])))
        
    if solve:
        results = sympy.solve_linear_system(system, *ynames)
    else:
        results = {}
    
    return results, (system, ynames, p_sa, p_sasr)

#%%
def bellman_sas(p_sa, p_sas, r_sas, gamma,
                s_names=None, a_names=None,
                solve=True):
    '''
    | 贝尔曼方程矩阵形式并求解，动力系统矩阵p_sas为P(s'|s,a)形式
    | p_sas为状态&动作-状态转移概率矩阵
    | r_sas为与p_sas一一对应的奖励表
    '''
    
    p_sa, p_sas = np.array(p_sa), np.array(p_sas)
    n_s, n_a = p_sa.shape
    n_sa, n_s = p_sas.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))
    
    p_sa = pd.DataFrame(p_sa, index=s_names, columns=a_names)
    p_sas = pd.DataFrame(p_sas, index=sa_names, columns=s_names)
    r_sas = pd.DataFrame(r_sas, index=sa_names, columns=s_names)
    
    part1 = np.eye(n_s)
    part2 = block_diag(*(p_sa.iloc[k, :] for k in range(p_sa.shape[0])))
    part12 = np.concatenate([part1, -part2], axis=1)
    
    part3 = -gamma * p_sas
    part4 = np.eye(n_sa)
    part34 = np.concatenate([part3, part4], axis=1)
    
    a_ = np.concatenate([part12, part34], axis=0)
    
    y1 = [[0] for _ in range(n_s)]
    y2 = p_sas * r_sas
    y2 = y2.sum(axis=1)
    y_ = np.concatenate((y1, pd.DataFrame(y2)), axis=0)
    
    system = np.concatenate([a_, y_], axis=1)
    system = sympy.Matrix(system)
    
    ynames = []
    for s in s_names:
        exec('v_{s} = symbols("v_{s}")'.format(s=s))
        ynames.append(eval('v_{}'.format(s)))
    for sa in sa_names:
        exec('v_{sa0}_{sa1} = symbols("v_{sa0}_{sa1}")'.format(sa0=sa[0], sa1=sa[1]))
        ynames.append(eval('v_{}_{}'.format(sa[0], sa[1])))
        
    if solve:
        results = sympy.solve_linear_system(system, *ynames)
    else:
        results = None
    
    return results, (system, ynames, p_sa, p_sas, r_sas)

#%%
def bellman_sas_sv(p_sa, p_sas, r_sas, gamma,
                   s_names=None, a_names=None,
                   solve=True):
    '''
    贝尔曼方程矩阵形式并求解，只考虑状态价值自身相互表示
    '''
    
    p_sa, p_sas = np.array(p_sa), np.array(p_sas)
    n_s, n_a = p_sa.shape
    n_sa, n_s = p_sas.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))
    
    p_sa = pd.DataFrame(p_sa, index=s_names, columns=a_names)
    p_sas = pd.DataFrame(p_sas, index=sa_names, columns=s_names)
    r_sas = pd.DataFrame(r_sas, index=sa_names, columns=s_names)
    
    d_sa = block_diag(*(p_sa.iloc[k, :] for k in range(p_sa.shape[0])))
    p_ss = np.matmul(d_sa, p_sas)
    p_ss.index = s_names
    
    i = pd.DataFrame(np.eye(p_ss.shape[0]),
                     index=p_ss.index,
                     columns=p_ss.columns)
    a_ = i - gamma * p_ss
    
    r_sas_ = p_sas * r_sas
    r_sa = r_sas_.sum(axis=1).to_frame()
    r_sa = r_sa.reset_index()
    r_sa.columns = ['sa', 'r']
    r_sa[['s', 'a']] = r_sa['sa'].apply(pd.Series)
    r_sa = df_rows2cols(r_sa[['s', 'a', 'r']], 'a', 'r')
    r_sa = r_sa.set_index('s')
    r_sa = r_sa.reindex(columns=a_names, index=s_names)
    
    r_s = (p_sa * r_sa).sum(axis=1).to_frame()
    r_s.columns = ['r_s']
    
    y_ = np.array(r_s['r_s'])
    
    system = np.concatenate([a_.values, y_.reshape(-1,1)], axis=1)
    system = sympy.Matrix(system)
    
    ynames = []
    for s in s_names:
        exec('v_{s} = symbols("v_{s}")'.format(s=s))
        ynames.append(eval('v_{}'.format(s)))
        
    if solve:
        results = sympy.solve_linear_system(system, *ynames)
    else:
        results = None
    
    return results, (system, ynames, p_ss, r_s, p_sa, p_sas, r_sas)

#%%
def bellman_sas_av(p_sa, p_sas, r_sas, gamma,
                   s_names=None, a_names=None,
                   solve=True):
    '''
    贝尔曼方程矩阵形式并求解，只考虑动作价值自身相互表示
    '''
    
    p_sa, p_sas = np.array(p_sa), np.array(p_sas)
    n_s, n_a = p_sa.shape
    n_sa, n_s = p_sas.shape
    
    if s_names is None:
        s_names = ['s%s'%k for k in range(1, n_s+1)]
    if a_names is None:
        a_names = ['a%s'%k for k in range(1, n_a+1)]        
    sa_names = list(product(s_names, a_names))
    
    p_sa = pd.DataFrame(p_sa, index=s_names, columns=a_names)
    p_sas = pd.DataFrame(p_sas, index=sa_names, columns=s_names)
    r_sas = pd.DataFrame(r_sas, index=sa_names, columns=s_names)
    
    i = pd.DataFrame(np.eye(p_sas.shape[0]),
                     index=p_sas.index,
                     columns=p_sas.index)
    
    p_sa_ = p_sa.copy()
    p_sa_.index.name = 's'
    p_sa_ = p_sa_.reset_index()
    p_sa_ = df_cols2rows(p_sa_, a_names, 'a', 'p')
    p_sa_['sa'] = p_sa_[['s', 'a']].apply(lambda x: (x['s'], x['a']), axis=1)
    p_sa_ = p_sa_.set_index('sa')[['p']]
    p_sa_ = p_sa_.transpose()
    p_sa_ = p_sa_.reindex(columns=p_sas.index)
    p_sa_ = pd.concat([p_sa_]*len(sa_names), axis=0)
    p_sa_.index = p_sa_.columns
    
    p_sasa = pd.concat([p_sas]*len(a_names), axis=1)
    p_sasa.columns = [(s, a) for a in a_names for s in p_sas.columns]
    p_sasa = p_sasa.reindex(index=p_sa_.index,
                            columns=p_sa_.columns)
    p_sasa = p_sasa * p_sa_
    
    a_ = i - gamma * p_sasa
    
    r_sas_ = p_sas * r_sas
    r_sa = r_sas_.sum(axis=1).to_frame()
    r_sa.columns = ['r_sa']
    
    y_ = np.array(r_sa['r_sa'])
    
    system = np.concatenate([a_.values, y_.reshape(-1,1)], axis=1)
    system = sympy.Matrix(system)
    
    ynames = []
    for sa in sa_names:
        exec('v_{sa0}_{sa1} = symbols("v_{sa0}_{sa1}")'.format(sa0=sa[0], sa1=sa[1]))
        ynames.append(eval('v_{}_{}'.format(sa[0], sa[1])))
        
    if solve:
        results = sympy.solve_linear_system(system, *ynames)
    else:
        results = None
    
    return results, (system, ynames, p_sasa, r_sa)

#%%
if __name__ == '__main__':
    from IPython.display import display
    from dramkit.gentools import TimeRecoder
    
    tr = TimeRecoder()
    # '''
    
    #%%
    # hungry game
    x, y = symbols('x y')
    alpha, beta, gamma = symbols('alpha beta gamma')
    # alpha, beta, gamma = 0.67, 0.75, 0.80
    # alpha, beta, gamma = 2/3, 3/4, 4/5
    # alpha, beta, gamma = symbols('0.67, 0.75, 0.80')
    # alpha, beta, gamma = symbols('2/3, 3/4, 4/5')
    
    s_names = ['Hgry', 'Full']
    a_names = ['notEat', 'Eat']
    
    # x, y = 1, 1
    p_sa = [[1-x,   x],
            [  y, 1-y]]
    
    # 1. 使用带奖励的状态&动作-状态&奖励转移矩阵
    r = [-3, -2, 1, 2]
    p_sasr = [[      0,    1, 0, 0, 0, 0,     0,      0],
              [1-alpha,    0, 0, 0, 0, 0, alpha,      0],
              [      0, beta, 0, 0, 0, 0,     0, 1-beta],
              [      0,    0, 0, 0, 0, 0,     1,      0]]
    
    res1, (sys1, ynames1, p_sa1, p_sasr1) = \
                bellman_sasr(p_sa, p_sasr, r, gamma,
                             s_names=s_names, a_names=a_names)
    display(res1)
    # for k, v in res1.items():
    #     display(k, v)
    
    
    # alpha_, beta_, gamma_ = 0.67, 0.75, 0.80
    alpha_, beta_, gamma_ = 2/3, 3/4, 4/5    
        
    # 2. 使用状态&动作-状态转移矩阵
    p_sas = [[      1,      0],
             [1-alpha,  alpha],
             [   beta, 1-beta],
             [      0,      1]]
    r_sas = [[-2, 0],
             [-3, 1],
             [-2, 2],
             [ 0, 1]]
    
    res2, (sys2, ynames2, p_sa2, p_sas2, r_sas2) = \
                    bellman_sas(p_sa, p_sas, r_sas, gamma,
                                s_names=s_names, a_names=a_names)
    # display(res2)
    for k, v in res2.items():
        display(k, v)
        display(k, v.subs([(alpha, alpha_), 
                           (beta, beta_),
                           (gamma, gamma_)]))
        
        
    # 3. 仅对状态价值自身相互表示
    res3, (sys3, ynames3, p_ss3, r_s3, p_sa3, p_sas3, r_sas3) = \
            bellman_sas_sv(p_sa, p_sas, r_sas, gamma,
                           s_names=s_names, a_names=a_names)
    res3_ = res3.copy()
    display(res3)
    for k, v in res3.items():        
        display(k, v)
        v_subs = v.subs([(alpha, alpha_), 
                         (beta, beta_),
                         (gamma, gamma_)])
        display(k, v_subs)
        
        
    # 4. 仅对动作价值自身相互表示
    res4, (sys4, ynames4, p_sasa4, r_sa4) = \
            bellman_sas_av(p_sa, p_sas, r_sas, gamma,
                           s_names=s_names, a_names=a_names)
    display(res4)
    for k, v in res4.items():        
        display(k, v)
        display(k, v.subs([(alpha, alpha_), 
                            (beta, beta_),
                            (gamma, gamma_)]))
        
        
    # 求状态价值最大值
    xv = yv = np.linspace(0, 1, 50)
    dfv = pd.DataFrame(product(xv, yv), columns=['x', 'y'])
    dfv['x'] = 1 - dfv['x']
    for k, v in res3_.items():
        v_name = str(k)
        dfv[v_name] = dfv[['x', 'y']].apply(
                    lambda xy: v.subs(
                        [(alpha, alpha_), (beta, beta_),
                         (gamma, gamma_),
                         (x, xy['x']), (y, xy['y'])]),
                    axis=1)
    print(dfv.max())
    # '''
    
    #%%
    tr.used()
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    